from ._lite_llm_client import (
  LiteLLMClient
)
from ._interfaces import (
  LLMMessage,
  LLMMessageRole,
	InferenceResult,
	InferenceOptions,
)
from ._config import (
  OpenAIModel,
  OpenAIConfig,
  GeminiModel,
  GeminiConfig,
  AnthropicModel,
  AnthropicConfig,
)



import os
from dotenv import load_dotenv
from enum import Enum
from typing import Optional, Union

from lite_llm_client._interfaces import LLMConfig

# load .env
load_dotenv()

###################################################
class OpenAIModel(Enum):
  GPT_4O = "gpt-4o"
  GPT_4O_MINI = "gpt-4o-mini"
  GPT_4_TURBO = "gpt-4-turbo"

class OpenAIConfig(LLMConfig):
  base_url: str
  api_key: str
  chat_completion_path: Optional[str] ="/v1/chat/completions"
  model:Union[OpenAIModel,str]

  def __init__(self,
               base_url:str="https://api.openai.com",
               api_key:str=None,
               model:Union[OpenAIModel,str]=OpenAIModel.GPT_4O):
    """
    parameters
    - api_key: if None, use environment variable "OPENAI_API_KEY"
    """

    self.base_url = base_url
    self.api_key = api_key
    if not self.api_key and "OPENAI_API_KEY" in os.environ:
      self.api_key = os.environ["OPENAI_API_KEY"]
    assert self.api_key and len(self.api_key) > 0, "api_key must be exists(check argument or environment variable)"
    self.model = model
  
  def get_chat_completion_url(self)->str:
    return f'{self.base_url}{self.chat_completion_path}'

###################################################
class AnthropicModel(Enum):
  CLAUDE_3_5_SONNET_20240620="claude-3-5-sonnet-20240620"
  CLAUDE_3_OPUS_20240229="claude-3-opus-20240229"

class AnthropicConfig(LLMConfig):
  base_url: str
  api_key: str
  chat_completion_path: Optional[str] ="/v1/messages"
  model:Union[AnthropicModel,str]

  max_tokens:int=1024

  def __init__(self,
               base_url:str="https://api.anthropic.com",
               api_key:str=None,
               model:Union[AnthropicModel,str]=AnthropicModel.CLAUDE_3_5_SONNET_20240620):
    """
    parameters
    - api_key: if None, use environment variable "ANTHROPIC_API_KEY"
    """

    self.base_url = base_url
    self.api_key = api_key
    if not self.api_key and "ANTHROPIC_API_KEY" in os.environ:
      self.api_key = os.environ["ANTHROPIC_API_KEY"]
    assert self.api_key and len(self.api_key) > 0, "api_key must be exists(check argument or environment variable)"
    self.model = model

  def get_chat_completion_url(self)->str:
    return f'{self.base_url}{self.chat_completion_path}'

###################################################
class GeminiModel(Enum):
  GEMINI_1_5_FLASH="gemini-1.5-flash"
  GEMINI_1_5_PRO="gemini-1.5-pro"

class GeminiConfig(LLMConfig):
  base_url: str
  api_key: str
  chat_completion_path: Optional[str] ="/v1/models"
  model:str

  max_tokens:int=1024

  def __init__(self,
               base_url:str="https://generativelanguage.googleapis.com",
               api_key:str=None,
               model:Union[GeminiModel,str]=GeminiModel.GEMINI_1_5_FLASH):
    """
    parameters
    - api_key: if None, use environment variable "GEMINI_API_KEY"
    """

    self.base_url = base_url
    self.api_key = api_key
    if not self.api_key and "GEMINI_API_KEY" in os.environ:
      self.api_key = os.environ["GEMINI_API_KEY"]
    
    assert self.api_key and len(self.api_key) > 0, "api_key must be exists(check argument or environment variable)"
    self.model = model if isinstance(model, str) else model.value

    #if not self.api_key:
    #  raise NotImplementedError()

  def get_chat_completion_url(self)->str:
    return f'{self.base_url}{self.chat_completion_path}/{self.model}'



from enum import Enum
import logging
from typing import Iterator, List
from lite_llm_client._config import OpenAIConfig
import requests

from lite_llm_client._interfaces import InferenceOptions, InferenceResult, LLMClient, LLMMessage, LLMMessageRole
from lite_llm_client._http_sse import SSEDataType, decode_sse

class OpenAIClient(LLMClient):
  config:OpenAIConfig

  def __init__(self, config:OpenAIConfig):
    self.config = config

    
  def _make_and_send_request(self, messages:List[LLMMessage], options:InferenceOptions, use_sse=False)->requests.Response:
    _options = options if options else InferenceOptions()
    msgs = []
    for msg in messages:
      role = None
      if msg.role == LLMMessageRole.USER:
        role = "user"
      elif msg.role == LLMMessageRole.SYSTEM:
        role = "system"
      elif msg.role == LLMMessageRole.ASSISTANT:
        role = "assistant"
      else:
        logging.fatal("unknown role")

      msgs.append({"role": role, "content": msg.content})

    model_name = self.config.model.value if isinstance(self.config.model, Enum) else self.config.model
    request = {
      "model": model_name,
      "messages": msgs,
      "temperature": _options.temperature,
    }

    if use_sse:
      request['stream'] = True

    #logging.info(f'request={request}')

    http_response = requests.api.post(
      self.config.get_chat_completion_url(),
      headers={'Content-Type': 'application/json', 'Authorization': f'Bearer {self.config.api_key}'},
      json=request
      )

    if http_response.status_code != 200:
      logging.fatal(f'response={http_response.text}')
      raise Exception(f'bad status_code: {http_response.status_code}')

    return http_response

  
  
  def _parse_response(self, inference_result:InferenceResult, response:dict):

    choices0 = response['choices'][0]
    inference_result.finish_reason = choices0['finish_reason']

    if 'usage' in response:
      usage = response['usage']
      inference_result.prompt_tokens = usage['prompt_tokens']
      inference_result.completion_tokens = usage['completion_tokens']
      inference_result.total_tokens = usage['total_tokens']

    logging.info(inference_result)

  def _parse_async_response(self, inference_result:InferenceResult, response:dict)->str:
    choices0 = response['choices'][0]
    finish_reason = choices0['finish_reason']
    if finish_reason is not None:
      inference_result.finish_reason = finish_reason

    delta = choices0['delta']
    if 'content' in delta:
      char = delta['content']
      inference_result.completion_tokens += len(char)
      inference_result.total_tokens += len(char)
      return char
    return None

  def async_chat_completions(self, messages:List[LLMMessage], options:InferenceOptions)->Iterator[str]:
    # TODO: count prompt length
    
    http_response = self._make_and_send_request(messages=messages, options=options, use_sse=True)

    for event in decode_sse(http_response, data_type=SSEDataType.JSON):
      """
      value example:
      {'id': 'chatcmpl-9qLv6AAbMZcZudyYUJ2SsSYGZs16y', 'object': 'chat.completion.chunk', 'created': 1722264344, 'model': 'gpt-4o-2024-05-13', 'system_fingerprint': 'fp_400f27fa1f', 'choices': [{'index': 0, 'delta': {'role': 'assistant', 'content': ''}, 'logprobs': None, 'finish_reason': None}]}
      """
      char = self._parse_async_response(inference_result=options.inference_result, response=event.event_value)

      if char:
        yield char
      else:
        # may be last?
        pass


  def chat_completions(self, messages:List[LLMMessage], options:InferenceOptions):
    http_response = self._make_and_send_request(messages=messages, options=options)
    response = http_response.json()
    #logging.info(f'response={response}')

    if options is not None and options.inference_result is not None:
      self._parse_response(options.inference_result, response)

    choices = response['choices']
    return choices[0]["message"]["content"]


import os
import sys
from typing import List, Union

sys.path.append(os.path.abspath('.'))
import logging
from _share import get_test_messages
from lite_llm_client import AnthropicConfig 
from lite_llm_client import AnthropicModel
from lite_llm_client import InferenceOptions
from lite_llm_client import LiteLLMClient

logging.basicConfig(level='debug')

def gen_instance()->LiteLLMClient:
  client = LiteLLMClient(AnthropicConfig(model=AnthropicModel.CLAUDE_3_OPUS_20240229))

  return client

def test_anthropic_sync():
  client = gen_instance()
  options = InferenceOptions()

  answer = client.chat_completions(messages=get_test_messages(), options=options)
  logging.info("{}".format(answer))
  logging.info(options.inference_result)


def test_anthropic_async():
  client = gen_instance()
  options = InferenceOptions()

  answer = client.async_chat_completions(messages=get_test_messages(), options=options)
  for a in answer:
    logging.info(a)
  logging.info(options.inference_result)




from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Iterator, List, Literal, Optional
from pydantic import BaseModel

@dataclass
class LLMConfig(ABC):
  @abstractmethod
  def get_chat_completion_url(self)->str:
    raise NotImplementedError

class LLMMessageRole(IntEnum):
  USER=1
  SYSTEM=2
  ASSISTANT=3

class LLMMessage(BaseModel):
  role: LLMMessageRole
  content: str

class InferenceResult(BaseModel):
  finish_reason:Literal['empty', 'stop']='empty'
  prompt_tokens:int=0
  completion_tokens:int=0
  total_tokens:int=0

class InferenceOptions(BaseModel):
  top_p:Optional[float]=None
  top_k:Optional[float]=None
  max_tokens: Optional[int]=None
  temperature:float=0.0
  inference_result:InferenceResult=InferenceResult()


class LLMClient(ABC):

  @abstractmethod
  def chat_completions(self, messages:List[LLMMessage], options:InferenceOptions)->str:
    raise NotImplementedError

  @abstractmethod
  def async_chat_completions(self, messages:List[LLMMessage], options:InferenceOptions)->Iterator[str]:
    raise NotImplementedError


from typing import Iterator, List
from lite_llm_client._anthropic_client import AnthropicClient
from lite_llm_client._config import GeminiConfig, LLMConfig, OpenAIConfig, AnthropicConfig
from lite_llm_client._gemini_client import GeminiClient
from lite_llm_client._interfaces import InferenceOptions, LLMClient, LLMMessage
from lite_llm_client._openai_client import OpenAIClient

class LiteLLMClient():
  """
  This is lite-llm-client class.
  it supports three types of client

  OpenAI usage:

  >>> from lite_llm_client import LiteLLMClient, OpenAIConfig
  >>> client = LiteLLMClient(OpenAIConfig(api_key="your api key"))

  Gemini usage:

  >>> from lite_llm_client import LiteLLMClient, GeminiConfig
  >>> client = LiteLLMClient(GeminiConfig(api_key="your api key"))

  Anthropic usage:

  >>> from lite_llm_client import LiteLLMClient, AnthropicConfig
  >>> client = LiteLLMClient(AnthropicConfig(api_key="your api key"))
  """
  config:LLMConfig
  client:LLMClient=None

  def __init__(self, config:LLMConfig):
    self.config = config

    if isinstance(config, OpenAIConfig):
      self.client = OpenAIClient(config)
    elif isinstance(config, AnthropicConfig):
      self.client = AnthropicClient(config)
    elif isinstance(config, GeminiConfig):
      self.client = GeminiClient(config)

    if not self.client:
      raise NotImplementedError()
    

  def chat_completions(self, messages:List[LLMMessage], options:InferenceOptions=None)->str:
    r"""chat completions function
    
    :param messages: messages
    :param options: (optional) options for chat completions
    :return answer of LLM

    """
    return self.client.chat_completions(messages=messages, options=options)

  def async_chat_completions(self, messages:List[LLMMessage], options:InferenceOptions=None)->Iterator[str]:
    r"""chat completions
    
    :param messages: messages
    :param options: (optional) options for chat completions
    :return parts of answer. use generator

    """
    return self.client.async_chat_completions(messages=messages, options=options)


from enum import Enum
import logging
from typing import Iterator, List

import requests
from lite_llm_client._config import AnthropicConfig
from lite_llm_client._http_sse import SSEDataType, decode_sse
from lite_llm_client._interfaces import InferenceOptions, InferenceResult, LLMClient, LLMMessage, LLMMessageRole


class AnthropicClient(LLMClient):
  config: AnthropicConfig
  
  def __init__(self, config:AnthropicConfig):
    self.config = config

  def _make_and_send_request(self, messages:List[LLMMessage], options:InferenceOptions, use_sse=False)->requests.Response:
    _options = options if options else InferenceOptions()
    msgs = []
    system_prompt = []
    for msg in messages:
      role = None
      if msg.role == LLMMessageRole.USER:
        role = "user"
      elif msg.role == LLMMessageRole.SYSTEM:
        system_prompt.append(msg.content)
        continue
      elif msg.role == LLMMessageRole.ASSISTANT:
        role = "assistant"
      else:
        logging.fatal("unknown role")

      msgs.append({"role": role, "content": msg.content})

    """
    https://docs.anthropic.com/en/api/messages
    
    system_prompt does not include messages.
    """
    model_name = self.config.model.value if isinstance(self.config.model, Enum) else self.config.model
    request = {
      "model": model_name,
      'max_tokens': self.config.max_tokens,
      "messages": msgs,
      "temperature": _options.temperature,
    }

    if len(system_prompt) > 0:
      request['system'] = "\n".join(system_prompt)

    if _options.top_k:
      request['top_k'] = _options.top_k
    if _options.top_p:
      request['top_p'] = _options.top_p

    if use_sse:
      request['stream'] = True

    http_response = requests.api.post(
      self.config.get_chat_completion_url(),
      headers={
        'Content-Type': 'application/json',
        'x-api-key': f'{self.config.api_key}',
        'anthropic-version': '2023-06-01',
        },
      json=request
      )

    if http_response.status_code != 200:
      logging.fatal(f'response={http_response.text}')
      raise Exception(f'bad status_code: {http_response.status_code}')

    return http_response

  def _parse_response(self, inference_result:InferenceResult, response:dict):
    reason_map = {
      'end_turn': 'stop'
    }

    if 'stop_reason' in response:
      stop_reason = response['stop_reason']
      inference_result.finish_reason = reason_map[stop_reason]

    if 'delta' in response:
      delta = response['delta']
      if 'stop_reason' in delta:
        stop_reason = delta['stop_reason']
        inference_result.finish_reason = reason_map[stop_reason]

    if 'usage' in response:
      usage = response['usage']

      if 'input_tokens' in usage:
        inference_result.prompt_tokens = usage['input_tokens']

      if 'output_tokens' in usage:
        inference_result.completion_tokens = usage['output_tokens']

      if inference_result.prompt_tokens > 0 and inference_result.completion_tokens > 0:
        inference_result.total_tokens = inference_result.prompt_tokens + inference_result.completion_tokens

  def async_chat_completions(self, messages:List[LLMMessage], options:InferenceOptions)->Iterator[str]:
    http_response = self._make_and_send_request(messages=messages, options=options, use_sse=True)

    for event in decode_sse(http_response, data_type=SSEDataType.JSON):
      #logging.info(event.event_value)
      self._parse_response(options.inference_result, event.event_value)

      """
      event_value example:
      {"type":"message_start","message":{"id":"msg_01C4SDTbBPX6yQiFSrgnY8jD","type":"message","role":"assistant","model":"claude-3-opus-20240229","content":[],"stop_reason":null,"stop_sequence":null,"usage":{"input_tokens":13,"output_tokens":3}}             }
      """
      event_type = event.event_value['type']
      if event_type == 'message_start':
        # just start signal
        continue

      if event_type == 'content_block_start':
        # ok start
        continue

      if event_type == 'ping':
        # what the ping
        continue

      if event_type == 'content_block_stop':
        continue
      if event_type == 'message_delta':
        """
        {"type":"message_delta","delta":{"stop_reason":"end_turn","stop_sequence":null},"usage":{"output_tokens":12}              }
        """
        continue
      if event_type == 'message_stop':
        continue

      if event_type == 'content_block_delta':
        """delta example:
        {"type":"content_block_delta","index":0,"delta":{"type":"text_delta","text":"Hello! How"} }
        """
        delta = event.event_value['delta']
        char = delta['text']
        yield char


  def chat_completions(self, messages:List[LLMMessage], options:InferenceOptions):
    http_response = self._make_and_send_request(messages=messages, options=options)
    response = http_response.json()

    #logging.info(response)
    if options is not None:
      self._parse_response(options.inference_result, response)

    content = response['content'][0]
    return content['text']


import logging
from typing import Iterator, List

import requests
from lite_llm_client._config import GeminiConfig
from lite_llm_client._http_sse import SSEDataType, decode_sse
from lite_llm_client._interfaces import InferenceResult, LLMClient, LLMMessageRole, LLMMessage, InferenceOptions, InferenceOptions, LLMMessage, LLMMessageRole


class GeminiClient(LLMClient):
  config: GeminiConfig
  
  def __init__(self, config:GeminiConfig):
    self.config = config

  def _make_and_send_request(self, messages:List[LLMMessage], options:InferenceOptions, use_sse=False)->requests.Response:
    _options = options if options else InferenceOptions(temperature=1.0, max_tokens=800, top_p=0.8, top_k=10)
    msgs = []
    system_prompt = []
    for msg in messages:
      role = None
      if msg.role == LLMMessageRole.USER:
        role = "user"
      elif msg.role == LLMMessageRole.SYSTEM:
        role = "system"
        continue
      elif msg.role == LLMMessageRole.ASSISTANT:
        role = "assistant"
      else:
        logging.fatal("unknown role")

      msgs.append({"role": role, "parts": [{'text':msg.content}]})

    
    generationConfig = {
      "temperature": _options.temperature,
      "maxOutputTokens": _options.max_tokens,
      "topP": _options.top_p,
      "topK": _options.top_k,
    }
    request = {
      "contents": msgs,
      "generationConfig": generationConfig,
    }

    if len(system_prompt) > 0:
      request['system'] = "\n".join(system_prompt)

    logging.info(f'request={request}')

    action = 'generateContent'
    alt = ''
    if use_sse:
      action = 'streamGenerateContent'
      alt = 'alt=sse&'
    url = f'{self.config.get_chat_completion_url()}:{action}?{alt}key={self.config.api_key}'
    http_response = requests.api.post(
      url,
      headers={
        'Content-Type': 'application/json',
        },
      json=request
      )

    if http_response.status_code != 200:
      logging.fatal(f'status_code={http_response.status_code} response={http_response.text}')
      raise Exception(f'bad status_code: {http_response.status_code}')

    return http_response

  def _parse_response(self, inference_result:InferenceResult, response:dict):
    reason_map = {
      'STOP': 'stop'
    }

    candidates0 = response['candidates'][0]

    finishReason = candidates0['finishReason']
    inference_result.finish_reason = reason_map[finishReason]

    usageMetadata = response['usageMetadata']
    inference_result.prompt_tokens = usageMetadata['promptTokenCount']
    inference_result.completion_tokens = usageMetadata['candidatesTokenCount']
    inference_result.total_tokens = usageMetadata['totalTokenCount']

  def async_chat_completions(self, messages:List[LLMMessage], options:InferenceOptions)->Iterator[str]:
    http_response = self._make_and_send_request(messages=messages, options=options, use_sse=True)

    for event in decode_sse(response=http_response, data_type=SSEDataType.JSON):

      """ value example:
      {"candidates": [{"content": {"parts": [{"text": " project's source code, organized into folders and files.\n* **Documentation:**  Some projects have a dedicated \"docs\" folder or a link to external documentation.\n* **Issues:**  This section lists any reported problems or feature"}],"role": "model"},"finishReason": "STOP","index": 0,"safetyRatings": [{"category": "HARM_CATEGORY_SEXUALLY_EXPLICIT","probability": "NEGLIGIBLE"},{"category": "HARM_CATEGORY_HATE_SPEECH","probability": "NEGLIGIBLE"},{"category": "HARM_CATEGORY_HARASSMENT","probability": "NEGLIGIBLE"},{"category": "HARM_CATEGORY_DANGEROUS_CONTENT","probability": "NEGLIGIBLE"}]}],"usageMetadata": {"promptTokenCount": 45,"candidatesTokenCount": 192,"totalTokenCount": 237}}
      """
      self._parse_response(options.inference_result, event.event_value)
      content = event.event_value['candidates'][0]['content']['parts'][0]
      char = content['text']
      yield char

  def chat_completions(self, messages:List[LLMMessage], options:InferenceOptions):
    http_response = self._make_and_send_request(messages=messages, options=options)
    response = http_response.json()
    logging.info(response)
    self._parse_response(options.inference_result, response)

    content = response['candidates'][0]['content']['parts'][0]
    return content['text']


from enum import Enum, auto
import json
import logging
from typing import Iterator, Union
from pydantic import BaseModel
from requests import Response

class SSEEvent(BaseModel):
  event_name:str
  event_value:Union[str,dict]

class SSEDataType(Enum):
  TEXT=auto()
  JSON=auto()

def _parse_sse(line:bytes):
  first_comma = line.find(b': ', 0)
  if first_comma == -1:
    raise ValueError(f'sse parse error (1)##{line}')

  name = line[:first_comma] 
  value = line[first_comma+2:]
  return name.decode(),value.decode()


def decode_sse(response:Response, data_type:SSEDataType, eoe:str='[DONE]')->Iterator[SSEEvent]:
  ct = response.headers.get('Content-Type') # Content-Type: text/event-stream; utf-8
  ct_values = ct.split(';')
  assert ct_values[0] == 'text/event-stream', f"response content-type does not 'text/event-stream' but '{ct_values[0]}'"

  #current_event = None
  for line in response.iter_lines(delimiter=b'\n'):
    if len(line) == 0:
      # SKIP empty line
      continue
    if line == b'\r':
      # for gemini.. maybe data delimiter is \r\n
      continue

    parsed_line = _parse_sse(line)

    if parsed_line[0] == 'event':
      #logging.debug(f'got event: {parsed_line[1]}')
      #current_event = parsed_line[0]
      continue

    if parsed_line[0] == 'data' and parsed_line[1] == eoe:
      #logging.debug("END OF EVENT")
      break

    #logging.debug(f'event({current_event}), value({parsed_line[1]})')

    value:str
    if SSEDataType.JSON == data_type:
      value = json.loads(parsed_line[1])
    #elif SSEDataType.TEXT == data_type:
    else:
      value = parsed_line[1]

    yield SSEEvent(event_name=parsed_line[0], event_value=value)
  


import os
import sys

sys.path.append(os.path.abspath('.'))
import logging
from _share import get_test_messages
from lite_llm_client import OpenAIConfig, OpenAIModel
from lite_llm_client import LiteLLMClient
from lite_llm_client import InferenceOptions

def gen_instance()->LiteLLMClient:
  client = LiteLLMClient(OpenAIConfig(
    model=OpenAIModel.GPT_4O_MINI
    ))

  return client

def test_oai_sync():
  client = gen_instance()
  options = InferenceOptions()

  answer = client.chat_completions(messages=get_test_messages(), options=options)
  logging.info("{}".format(answer))
  logging.info(options.inference_result)

def test_oai_async():
  client = gen_instance()
  options = InferenceOptions()

  answer = client.async_chat_completions(messages=get_test_messages(), options=options)
  for a in answer:
    logging.debug(a)

  logging.info(options.inference_result)
  

